# -*- coding: utf-8 -*-

import warnings
warnings.warn("pytesmo.timdate.dekad has moved to the cadati pip package "
              "Import from pytesmo will not work in future versions of the package",
              DeprecationWarning)

from cadati.dekad import *